#!/usr/bin/perl

# This script runs nmap, a portscanner, against a specified subnet
# and reports any instances of unknown open ports on hosts.  Known
# ports on hosts can be enumerated in the localscan.conf file and
# are not reported by localscan.
#
# Copyright (C) 1999 by Dylan Greene  
# Last modified Wednesday, 19 January 2000

# Added support for multiple space-delimited subnets from either the
# command-line or localscan.conf.  If a subnet(s) is specified on the
# command-line, it supersedes any subnet(s) defined in localscan.conf.

# Define a path variable (to be filled in by make_conf.pl).
$path;

unless ( $path ) {
    # Check to see if make_conf.pl was run--need to set the $path variable.
    die "\nYou must run make_conf.pl once before running localscan.pl\n\n";
}

$scan_report = join('',$path,"/scan_report");
$SCAN_REPORT = join('',">",$path,"/scan_report");
$CONF_FILE = join('',"<",$path,"/localscan.conf");

# First, let's create a temp file which will store any output generated
#open(SCAN_REPORT, ">scan_report");
open SCAN_REPORT or die "Can't open scan_report: $!\n";

# Put a header in that file
printf SCAN_REPORT "The following unrecognized services were found:\n";

# Read in the information contained in localscan.conf.
#open(CONF_FILE, "<localscan.conf") or die "Localscan.conf doesn't exist!";
open CONF_FILE or die "Can't open conf_file: $!\n";

@conf_items = <CONF_FILE>;
close(CONF_FILE);

# Figure out what each item read from localscan.conf is.
foreach my $item ( @conf_items ) {

    # Kill off any leading whitespace
    $item =~ s/^\s+//;
    chomp $item;
    if ( $item =~ /^#/ ) {

        # Is the item a comment?
	# Do nothing . . .

    } elsif ( $item eq "" ) {

	# Is the item a blank line?
	# Do nothing . . .

    } elsif ( $item =~ /^subnet / ) {

	# Is it the "subnet" field?
	$subnet_from_conf = $item;
	$subnet_from_conf =~ s/^subnet //;

    } elsif ( $item =~ /^mailto / ) {

	# Is it the "mailto" field?
	$mailto = $item;
	$mailto =~ s/^mailto //;

    } elsif ( $item =~ /^allclr / ) {

	# Is it the "allclr" field?
	$allclr = $item;
	$allclr =~ s/^allclr //;

    } elsif ( $item =~ /^nmap / ) {

	# Is the item the path to nmap?
	$nmap_path = $item;
	$nmap_path =~ s/^nmap //;

    } elsif ( $item =~ /^ignore / ) {

	# It'd better be an ignore item, then!
	# Split the item up and store it in a hash.
	# If the key (host) exists in the hash already, append
	# the next port(s) in a space-delimited list.

	my @split_item = split(/\s+/, $item);

	# Get rid of the first element--it's just the word "ignore"
	shift @split_item;

	# Get the second (now the first) element, the hostname.
	my $host = shift @split_item;

	# Now all that's left are the ports, so
	my $ports = join(' ', @split_item);

	# Just for kicks, assume that someone edited the localscan.conf
	# file by hand and added a duplicate line.  The next if() keeps
	# the hash from overwriting it.  There should be no problems with
	# duplicate port numbers inside the hash value, if that happens.

	if ( $ignore{$host} ) {
	    $ignore{$host} = join(' ', $ignore{$host}, $ports);
	} else {
	    $ignore{$host} = $ports;
	}

    } else {

	# If $item matches none of the above conditions, we have a
	# badly-formatted file

#	open STDERR, ">&STDERR"; 
	print STDERR "Incorrectly formatted localscan.conf file: $item";
#	close STDERR;

	# But be tolerant, don't kill the script.  Maybe the rest of
	# it'll work fine.

	# Still, though, make sure the admin knows about it.
	print SCAN_REPORT "Incorrectly formatted localscan.conf file: $item";

	# And set a warning flag so that the script knows to send mail
	$warning = 1;
    }
}


chomp %ignore;

# At this point we should have an hash of ignore conditions called
# %ignore, keyed with the hostname or IP.  The value of each entry
# is the port or ports (seperated by spaces) to ignore.

# Now call nmap with the $subnet variable (taken from the localscan.conf
# file or from the command-line) and store the output in an  array for processing.

# Check for command-line arguments
if ( $ARGV[0] ) {
    # Pass the arguments in turn through a foreach loop

    foreach $subnet ( @ARGV ) {

	chomp $subnet;
	nmap_processing();

    }
    $scanned_subnets = join(' ',@ARGV);

} else {
    # Pass the possibly space-delimited elements of $subnet through
    # foreach loop, too

    foreach $subnet ( split(/\s+/,$subnet_from_conf) ) {
	
	chomp $subnet;
	nmap_processing();

    }
    $scanned_subnets = $subnet_from_conf;
}

sub nmap_processing {

    $argument = join('',$nmap_path,"/nmap -sT '",$subnet,"'");
    chomp $argument;
    @data = `$argument`;

    # Now for the actual meat of the script--processing the data returned
    # from the nmap call.
    
    foreach my $line ( @data ) {
	chomp $line;
	    
	if ( $line =~ /^Interesting/ ) {
	    
	    # grab the IP address of the host
	    @host_info = split(/\s+/, $line);

	    if ( $host_info[4] ) {

		# Looks like we've got DNS info
		$ip = $host_info[4];
		$hostname = $host_info[3];
		chomp $hostname;

	    } else {

		# No DNS info here!
		$ip = $host_info[3];

	    }

	    $ip =~ s/\(//;
	    $ip =~ s/\)//;
	    $ip =~ s/://;
	    chomp $ip;
	    
	}
	if ( $line =~ /^[0-9]/ ) {

	    # Snap up the ports listed as open,
	    @port = split(/\s+/, $line);
	    
	    # $open_numeric is the port number, $open_name is the name of
	    # the service on that port.
	    $open_numeric = $port[0];
	    $open_name = $port[3];

	    # Now, let's do some checking.  If $ip is a key in our hash of
	    # ignore conditions, then it's possible that the port returned
	    # is also there.

	    # Use a flag to keep track of whether a port has matched or not
	    $ignore_flag = 0;

	    if ( $ignore{$ip} ) {
		
		# If $open_numeric or $open_name match any of the ports
		# listed as values in $ignore{$ip}, set the ignore_flag to 1.
		    
		foreach my $ignore_port ( split( /\s+/, $ignore{$ip} ) ) { 
		    
		    if ( $ignore_port eq "ALL" ) {
			
			# The "ALL" option was specified in localscan.conf
			# Therefore, set the ignore flag
			$ignore_flag = 1;

		    } elsif ( ( $ignore_port eq $open_numeric ) || ( $ignore_port eq $open_name) ) {

			# It matches, so the port currently being examined ($open_numeric)
			# is not to be reported on.
			
			# Therefore, set the ignore flag
			$ignore_flag = 1;
			
		    }
		}
	    }

	    if ( $ignore_flag == 0 ) {
		
		# This port on this host didn't match any of the
		# exceptions in the %ignore hash.  Report it!
		
		# Write the portname/number/ip/hostname combo to the 
		# temp output file--properly formatted, of course.  :)
		
		$output = join(' ',$hostname,$ip,"port",$open_numeric,$open_name,"\n");
		print SCAN_REPORT $output;
	    
		# Stick a counter in here to allow us to keep track
		# of how many exceptions are detected.
		$counter++;
		
	    }
	}	 
    }
}



# Now that the scan has been filtered, mail any output to the 
# $mailto address.

if ( $counter || $warning ) {

    # Either an unknown service was detected, or an error
    # occurred while processing the localscan.conf file.
    # In either case, mail needs to be sent.

    # Write the counter to the report file
    print SCAN_REPORT "*** $counter exceptions";

    # Close the report file
    close(SCAN_REPORT);

    # Mail the report file to the $mailto address
    `cat $scan_report | mail -s \"Localscan results for $scanned_subnets\" $mailto`;

} elsif ( $allclr eq "yes" ) {

    # Nothing unknown was detected, but the allclr flag was set,
    # which means that the admin wants to /know/ that everything's
    # okay.  So mail needs to be sent.

    `echo \"No dangerous services detected\" | mail -s \"Localscan results for $scanned_subnets\" $mailto`;

    # Close the report file
    close(SCAN_REPORT);

} else {

    # Nothing happened, but we still need to close the report file  :)
    close(SCAN_REPORT);

}

# Finally, delete the report file
unlink($scan_report);





